/*
 * INET		Port Base VLAN
 *		Ethernet-type device handling.
 */

#include <asm/uaccess.h> /* for copy_from_user */
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <net/datalink.h>
#include <linux/mm.h>
#include <linux/in.h>
#include <linux/init.h>
#include <net/arp.h>
#include <linux/rtnetlink.h>
#include <linux/brlock.h>
#include <linux/notifier.h>
#include "pbvlan.h"
#include "cpmac_debug.h"

/* Global VLAN variables */
struct net_device *pb_vlan_list[MAX_PB_VLAN]={NULL,NULL,NULL,NULL};
/* Switch vlan table */
unsigned short pbvlan_ingress[MAX_PB_VLAN]={0x3e,0x3d,0x3b,0x37};
/* Real port map to vlan IF */
int pbvlan_group[MAX_PB_VLAN]={3,3,3,3};

void reset_ingress( void )
{
	//reset ingress value
	pbvlan_ingress[0]=0x3e;
	pbvlan_ingress[1]=0x3d;
	pbvlan_ingress[2]=0x3b;
	pbvlan_ingress[3]=0x37;

}


static int vlan_device_event(struct notifier_block *, unsigned long, void *);

struct notifier_block vlan_notifier_block = {
	notifier_call: vlan_device_event,
};




/* End of global variables definitions. */

/*
 * Function vlan_proto_init (pro)
 *
 *    Initialize VLAN protocol layer, 
 *
 */
void pbvlan_proto_init(void)
{
	/* Register us to receive netdevice events */
	register_netdevice_notifier(&vlan_notifier_block);
}

/*
 *     Module 'remove' entry point.
 *     o delete /proc/net/router directory and static entries.
 */ 
void pbvlan_cleanup(void)
{
	/* Un-register us from receiving netdevice events */
	unregister_netdevice_notifier(&vlan_notifier_block);

}


/** This method will explicitly do a dev_put on the device if do_dev_put
 * is TRUE.  This gets around a difficulty with reference counting, and
 * the unregister-by-name (below).  If do_locks is true, it will grab
 * a lock before un-registering.  If do_locks is false, it is assumed that
 * the lock has already been grabbed externally...  --Ben
 */
int unregister_pb_vlan_dev(int real_dev_ifindex, unsigned short vlan_id,
			       int do_dev_put, int do_locks)
{
	struct net_device *dev = NULL;

	/* sanity check */
	if ((vlan_id >= MAX_PB_VLAN))
		return -EINVAL;

	dev = pb_vlan_list[vlan_id];

	if (dev) {
		/* Take it out of our own structures */
		pb_vlan_list[vlan_id] = NULL;


		if (do_dev_put)
			dev_put(dev);

		/* TODO:  Please review this code. */
		if (do_locks) {
			rtnl_lock();
			unregister_netdevice(dev);
			rtnl_unlock();
		} else {
			unregister_netdevice(dev);
		}

		MOD_DEC_USE_COUNT;
	}
        
        return 0;
}

int unregister_pb_vlan_device(const char *vlan_IF_name)
{
	struct net_device *dev = NULL;


	dev = dev_get_by_name(vlan_IF_name);
	if (dev) {
		if (dev->priv_flags & IFF_PB_VLAN) {
			return unregister_pb_vlan_dev(
				PBVLAN_DEV_INFO(dev)->real_dev->ifindex,
				(unsigned short)(PBVLAN_DEV_INFO(dev)->vlan_id),
				1 /* do dev_put */, 1 /* do locking */);
		} else {
			dev_put(dev);
			return -EPERM;
		}
	} else {
		return -EINVAL;
	}
}

/*  Attach a VLAN device to a mac address (ie Ethernet Card).
 *  Returns the device that was created, or NULL if there was
 *  an error of some kind.
 */
struct net_device *register_pb_vlan_device(const char* eth_IF_name,
					       unsigned short VLAN_ID)
{
	struct net_device *new_dev;
	struct net_device *real_dev; /* the ethernet device */
	int malloc_size = 0;


	if (VLAN_ID >= 0xfff || VLAN_ID > MAX_PB_VLAN || pb_vlan_list[VLAN_ID])
		goto out_ret_null;

	/* find the device relating to eth_IF_name. */
	real_dev = dev_get_by_name(eth_IF_name);
	if (!real_dev)
		goto out_ret_null;


	malloc_size = (sizeof(struct net_device));
	new_dev = (struct net_device *) kmalloc(malloc_size, GFP_KERNEL);

	if (new_dev == NULL)
		goto out_put_dev;

	pb_vlan_list[VLAN_ID] = new_dev;

	memset(new_dev, 0, malloc_size);

	/* set us up to not use a Qdisc, as the underlying Hardware device
	 * can do all the queueing we could want.
	 */
	/* new_dev->qdisc_sleeping = &noqueue_qdisc;   Not needed it seems. */
	new_dev->tx_queue_len = 0; /* This should effectively give us no queue. */

	/* Put our vlan.VID in the name.
	 * Name will look like:	 vlan5
	 */
	sprintf(new_dev->name, "vlan%i", VLAN_ID);
		    
	/* set up method calls */
	new_dev->init = vlan_dev_init;
	new_dev->features |= NETIF_F_DYNALLOC ; 
	    
	/* new_dev->ifindex = 0;  it will be set when added to
	 * the global list.
	 * iflink is set as well.
	 */
	new_dev->get_stats = vlan_dev_get_stats;
	    
	/* IFF_BROADCAST|IFF_MULTICAST; ??? */
	new_dev->flags = real_dev->flags;
	new_dev->flags &= ~IFF_UP;

	/* Make this thing known as a VLAN device */
	new_dev->priv_flags |= IFF_PB_VLAN;
				
	/* need 4 bytes for extra VLAN header info,
	 * hope the underlying device can handle it.
	 */
	new_dev->mtu = real_dev->mtu;
	new_dev->change_mtu = vlan_dev_change_mtu;

	/* TODO: maybe just assign it to be ETHERNET? */
	new_dev->type = real_dev->type;

	new_dev->hard_header_len = real_dev->hard_header_len;

	new_dev->priv = kmalloc(sizeof(struct pbvlan_dev_info),
				GFP_KERNEL);
	    
	if (new_dev->priv == NULL) {
		kfree(new_dev);
		goto out_put_dev;
	}

	memset(new_dev->priv, 0, sizeof(struct pbvlan_dev_info));

	memcpy(new_dev->broadcast, real_dev->broadcast, real_dev->addr_len);
	memcpy(new_dev->dev_addr, real_dev->dev_addr, real_dev->addr_len);
	new_dev->addr_len = real_dev->addr_len;

	new_dev->open = vlan_dev_open;
	new_dev->stop = vlan_dev_stop;
	new_dev->hard_header = real_dev->hard_header;

	new_dev->hard_start_xmit = vlan_dev_hard_start_xmit;
	//new_dev->rebuild_header = real_dev->rebuild_header;
	//new_dev->hard_header_parse = real_dev->hard_header_parse;
	//new_dev->set_mac_address = real_dev->set_mac_address;
	//new_dev->set_multicast_list = real_dev->set_multicast_list;

	PBVLAN_DEV_INFO(new_dev)->vlan_id = VLAN_ID; /* 1 through 0xFFF */
	PBVLAN_DEV_INFO(new_dev)->real_dev = real_dev;

	/* TODO: Please check this: RTNL   --Ben */
	rtnl_lock();
	register_netdevice(new_dev);
	rtnl_unlock();
	    
	/* NOTE:  We have a reference to the real device,
	 * so hold on to the reference.
	 */
	MOD_INC_USE_COUNT; /* Add was a success!! */
	printk("Allocated new device successfully, returning.\n");
	return new_dev;

out_put_dev:
	dev_put(real_dev);

out_ret_null:
	return NULL;
}

static int vlan_device_event(struct notifier_block *unused, unsigned long event, void *ptr)
{
	struct net_device *dev = (struct net_device *)(ptr);
	struct net_device *vlandev = NULL;
	int i = 0;

	switch (event) {
	case NETDEV_CHANGEADDR:
		/* Ignore for now */
		break;

	case NETDEV_GOING_DOWN:
		/* Ignore for now */
		break;

	case NETDEV_DOWN:
		/* TODO:  Please review this code. */
		/* put all related VLANs in the down state too. */
		{
			int flgs = 0;
			for (i = 0; i < MAX_PB_VLAN; i++) {
				vlandev = pb_vlan_list[i];
				if (!vlandev ||
				    (PBVLAN_DEV_INFO(vlandev)->real_dev != dev) ||
				    (!(vlandev->flags & IFF_UP)))
					continue;
	
				flgs = vlandev->flags;
				flgs &= ~IFF_UP;
				dev_change_flags(vlandev, flgs);
			}
		}
		break;

	case NETDEV_UP:
		{
			int flgs = 0;
			for (i = 0; i < MAX_PB_VLAN; i++) {
				vlandev = pb_vlan_list[i];
				if (!vlandev ||
				    (PBVLAN_DEV_INFO(vlandev)->real_dev != dev) ||
				    (!(vlandev->flags & IFF_UP)))
					continue;
	
				flgs = vlandev->flags;
				flgs |= IFF_UP;
				dev_change_flags(vlandev, flgs);
			}
		}
		break;
		
	case NETDEV_UNREGISTER:
		/* Ignore for now */
		break;
	};

	return NOTIFY_DONE;
}



void unregister_pb_vlan_device_all( void )
{
	int vlan_id;
	char vlan_name[16]="";

	dbg("---> unregister_vlan()\n");

	for( vlan_id =0; vlan_id < MAX_PB_VLAN; vlan_id ++)
	{
		if(  pb_vlan_list[ vlan_id] == NULL ) 
			continue;

		sprintf(vlan_name,"vlan%d",vlan_id);
		unregister_pb_vlan_device(vlan_name);
		dbg("   unregister_pb_vlan_device( %s )\n", vlan_name);
	}

	pbvlan_cleanup();
	
}


void register_pb_vlan_device_all( void )
{
	int port_id;
	int vlan_id;

	for ( port_id =0 ; port_id < MAX_PB_VLAN ; port_id ++)
	{
		vlan_id =  pbvlan_group[port_id];
		if(  pb_vlan_list[ vlan_id] != NULL)
			continue;

		register_pb_vlan_device( ETH_IF_NAME, vlan_id);		
		dbg("   register_pb_vlan_device( %s, %d )\n", ETH_IF_NAME, vlan_id);
	}

	pbvlan_proto_init();
}


/*


	return:
		the vlan device for the lan_port
		NULL if vlan is disable
*/
struct net_device * pb_vlan_get_device( int lan_port_id)
{
	int vlan_id;

	if (lan_port_id >= MAX_PB_VLAN || lan_port_id < 0)
	{
		dbg("--->pb_vlan_get_device() port=%d, error\n", lan_port_id);
		return NULL;
	}

	vlan_id = pbvlan_group[lan_port_id];


	dbg("--->pb_vlan_get_device() port=%d, vlan=%d, dev=%x\n", lan_port_id, vlan_id, (unsigned int) pb_vlan_list[vlan_id]);

	return pb_vlan_list[vlan_id];
}


/*
	get Destination Port Vector for a vlan

	see 88E6060 data sheet
	3.5.4 switch's ingress trailer 

*/
unsigned short pb_vlan_get_DPV( int vlan_id)
{
	int lan_port_id;
	unsigned short ret =0;	

	dbg("---> pb_vlan_get_DPV() ");

	for ( lan_port_id =0; lan_port_id < MAX_PB_VLAN ; lan_port_id++)
	{
		if ( pbvlan_group[ lan_port_id] == vlan_id)
		{
			ret = (pbvlan_ingress[lan_port_id] & 0x0F) | (1<<lan_port_id);

			dbg(" lan_port_id=%d ", lan_port_id);
			break;
		}
				
	}

	dbg(" ret=%x\n", ret);
	return ret;
}

